uniform bool colorMapping;
uniform sampler2D colorMap;

uniform bool bumpMapping;
uniform sampler2D normalMap;

uniform float scale;

uniform bool lighting;

uniform bool absorption;
uniform vec3 planePoint;
uniform vec4 planeColor;
uniform vec3 planeNormal;

attribute vec3 tangent;

uniform bool diffraction;

varying float a;
varying vec2 uv;
varying vec3 lightVec;
varying vec3 eyeVec;

vec3 lambda2rgb(float lambda)
{
    const float ultraviolet = 400.0;
    const float infrared    = 700.0;

    // map visible wavelength range to 0.0 -> 1.0
    float a = (lambda-ultraviolet) / (infrared-ultraviolet);

    // bump function for a quick/simple rainbow map
    const float C = 7.0;        // controls width of bump
    vec3 b = vec3(a) - vec3(0.75, 0.5, 0.25);
    return max((1.0 - C * b * b), 0.0);
}

void main(void)
{
	a=0.0;

	gl_Position = ftransform();
	gl_ClipVertex = gl_ModelViewMatrix * gl_Vertex;

	uv = vec2(gl_MultiTexCoord0.x*scale, gl_MultiTexCoord0.y*scale);

	if(absorption) {
			// work out the varying plane absorption

			vec4	ecPosition = gl_ModelViewMatrix * gl_Vertex;
			vec3	ecPosition3 = (vec3 (ecPosition)) / ecPosition.w;

			vec3	P1=ecPosition.xyz/ecPosition.w;				// vertex in world coords
			float	b=dot(planeNormal, (-P1));
			
			if(b>0.001) {
				float A=dot(planeNormal, (planePoint-P1));
				a=A/b;
			}

			a=a*length(P1);
			a=a/5000.0;
			a=clamp(a, 0.0, 1.0);
	}

	vec4	diffract=vec4(0.0, 0.0, 0.0, 0.0);

	if(diffraction) {
		vec4		ecPosition = gl_ModelViewMatrix * gl_Vertex;
		vec3		ecPosition3 = (vec3 (ecPosition)) / ecPosition.w;
		vec3		VP = vec3 (gl_LightSource[0].position) - ecPosition3;
		vec3		halfVector = normalize(VP + vec3 (0.0, 0.0, 1.0));

		vec3		T = gl_NormalMatrix * tangent;
		float		u = abs(dot(T, halfVector));
		vec3		diffColor = vec3(0.0);

		const int 	numSpectralOrders = 3;
		const float 	GratingSpacing=5000.0;
		const float 	SurfaceRoughness=10.0;

		for (int m = 1; m <= numSpectralOrders; ++m) {
			float lambda = GratingSpacing * u / float(m);
			diffColor += lambda2rgb(lambda);
		}

		// add specular highlight
/*
		float w = dot(normal, halfVector);
		float e = SurfaceRoughness * u / w;
		vec3 hilight = vec3(gl_FrontMaterial.specular*exp(-e * e));

		diffColor=clamp(diffColor+hilight, 0.0, 1.0);
*/
		diffract=vec4(diffColor.r, diffColor.g, diffColor.b, 1.0);
	}

	if(bumpMapping) {
		// set things up for per-fragment bump-mapped lighting

		vec3 n = normalize(gl_NormalMatrix * gl_Normal);
		vec3 t = normalize(gl_NormalMatrix * tangent);
		vec3 b = cross(n, t);

		vec3 vVertex = vec3(gl_ModelViewMatrix * gl_Vertex);
		vec3 tmpVec = gl_LightSource[0].position.xyz - vVertex;

		lightVec.x = dot(tmpVec, t);
		lightVec.y = dot(tmpVec, b);
		lightVec.z = dot(tmpVec, n);

		tmpVec = -vVertex;
		eyeVec.x = dot(tmpVec, t);
		eyeVec.y = dot(tmpVec, b);
		eyeVec.z = dot(tmpVec, n);

		// gl_FrontColor = mix(diffract, vec4(1.0), 0.5);

		return;
	}

	if(a>=1.0) {
		gl_FrontColor = planeColor;
		return;
	}

	if(!lighting) {
		gl_FrontColor = gl_Color;
		return;
	}

	// standard per-vertex lighting

	vec4	ecPosition = gl_ModelViewMatrix * gl_Vertex;
	vec3	ecPosition3 = (vec3 (ecPosition)) / ecPosition.w;
	vec3	normal = normalize(gl_NormalMatrix * gl_Normal);
	vec3	VP = vec3 (gl_LightSource[0].position) - ecPosition3;
	float	d = length(VP);

	float	attenuation = 1.0 / (	gl_LightSource[0].constantAttenuation +
  					gl_LightSource[0].linearAttenuation * d +
      					gl_LightSource[0].quadraticAttenuation * d * d);

	vec3	halfVector = normalize(VP + vec3 (0.0, 0.0, 1.0));
	float	nDotVP = max(0.0, dot(normal, normalize(VP)));

	vec4	ambient_light  = gl_LightSource[0].ambient * attenuation;
	vec4	diffuse_light  = gl_LightSource[0].diffuse * nDotVP * attenuation;
	vec4	color = gl_FrontLightModelProduct.sceneColor +
			gl_FrontMaterial.ambient*ambient_light+
			gl_FrontMaterial.diffuse*diffuse_light;


	if(diffraction) {
		// use diffraction "specular"

		//color = mix(color, diffract, 0.2);
		color+=diffract;
	}
	else {
		// work out regular specular

		float nDotHV = max(0.0, dot(normal, halfVector));
		float pf=0.0;

		if(nDotVP != 0.0) {
			pf = pow(nDotHV, gl_FrontMaterial.shininess);

			vec4	specular_light = gl_LightSource[0].specular * pf * attenuation;

			color+=gl_FrontMaterial.specular*specular_light;
		}
	}

	color = clamp(color, 0.0, 1.0);

	if(diffraction) {
	}

	gl_FrontColor=color;
}